import configparser
import os
import warnings
from os.path import basename, splitext
from urllib.parse import urlparse

import requests
from rich.console import Console
from sqlalchemy import inspect, text
from sqlalchemy.exc import DBAPIError

from piperider_cli.error import PipeRiderConnectorError, AwsCredentialsError, AwsUnExistedS3Bucket, \
    PipeRiderDataBaseConnectionError, PipeRiderDataBaseEncodingError
from . import DataSource
from .field import PathField, TextField

AWS_CREDENTIAL_PATH = os.path.expanduser('~/.aws/credentials')


def _parquet_path_validate_func(answer, current) -> bool:
    if current is None or current == '':
        return False

    if current.startswith('http://') or current.startswith('https://'):
        try:
            # Check header only, don't download entire file
            return requests.head(current).status_code < 400
        except Exception as e:
            console = Console()
            console.print(f'\n  [red]Invalid http request[/red]: {str(e)}')
        return False

    if current.startswith('s3://'):
        return True

    # File path
    print(current)
    return os.path.exists(current)


def get_s3_bucket_region(s3_bucket):
    try:
        s3_url = f'https://{s3_bucket}.s3.amazonaws.com/'
        response = requests.get(s3_url)
        if response.status_code == 403:
            return response.headers.get('x-amz-bucket-region')
        else:
            return ''
    except Exception:
        return ''


def get_aws_credentials():
    if os.path.exists(AWS_CREDENTIAL_PATH):
        config = configparser.RawConfigParser()
        config.read(AWS_CREDENTIAL_PATH)
        profiles = config.sections()
        if len(profiles) == 0:
            return None, None
        return config.get(profiles[0], 'aws_access_key_id'), config.get(profiles[0], 'aws_secret_access_key')

    return os.getenv('AWS_ACCESS_KEY_ID'), os.getenv('AWS_SECRET_ACCESS_KEY')


class DuckDBDataSource(DataSource):

    def __init__(self, name, **kwargs):
        super().__init__(name, 'duckdb', **kwargs)

        self.credential_source = 'config'
        self.fields = [
            PathField('path', description='Path of database file'),
            TextField('schema', description='Name of the schema', optional=True, default='main'),
        ]

        warnings.filterwarnings('ignore', message="duckdb-engine doesn't yet support reflection on indices")
        self.legacy_engine = None

    def validate(self):
        return self._validate_required_fields()

    def verify_connector(self):
        try:
            import duckdb
            import duckdb_engine
            import chardet
        except Exception as e:
            return PipeRiderConnectorError(str(e), self.type_name)

    def to_database_url(self, database):
        credential = self.credential
        from piperider_cli.configuration import FileSystem
        dbpath = credential.get('path')
        if os.path.isabs(dbpath) is False:
            current_directory_path = os.path.join(os.getcwd(), dbpath)
            working_directory_path = os.path.join(FileSystem.WORKING_DIRECTORY, dbpath)
            if os.path.exists(current_directory_path):
                duckdb_path = os.path.abspath(current_directory_path)
            else:
                duckdb_path = os.path.abspath(working_directory_path)
        else:
            duckdb_path = dbpath
        if not os.path.exists(duckdb_path):
            raise PipeRiderDataBaseConnectionError(self.name, self.type_name, db_path=duckdb_path)
        return f"duckdb:///{duckdb_path}"

    def _formalize_table_name(self, name):
        if not name:
            return f'{self.type_name}_table'

        if not name[0].isalpha():
            name = 'TABLE_' + name

        if len(name) > 120:
            name = name[:115] + '__'

        return name.replace('.', '_')

    def _get_display_description(self):
        cred = self.credential
        return f"type={self.type_name}, dbpath={cred.get('path')}, schema={cred.get('schema')}"

    def get_database(self):
        engine = self.get_engine_by_database()
        with engine.connect() as conn:
            result = conn.execute(text('pragma database_list')).fetchone()
            # 'pragma database_list' returns a tuple with three columns
            if result and len(result) == 3:
                seq, name, file = result
                return name
        return 'main'

    def is_legacy_duckdb(self):
        # NOTE: the old duckdb cannot open multi-instance to a single file
        import duckdb
        from packaging import version as v
        is_legacy = v.parse(duckdb.__version__) < v.parse('0.8.1')
        return is_legacy

    def get_engine_by_database(self, database=None):
        if self.is_legacy_duckdb():
            if self.legacy_engine is None:
                self.legacy_engine = self.create_engine(database)
            return self.legacy_engine

        return super().get_engine_by_database(database)

    @property
    def threads(self):
        if self.is_legacy_duckdb():
            return 1
        return super().threads

    def get_schema(self):
        cred = self.credential
        schema = cred.get('schema')
        if schema is None:
            schema = inspect(self.get_engine_by_database()).default_schema_name
        return schema


class CsvDataSource(DuckDBDataSource):
    def __init__(self, name, **kwargs):
        super(DuckDBDataSource, self).__init__(name, 'csv', **kwargs)

        self.credential_source = 'config'
        self.fields = [
            PathField('path', description='Path of csv file'),
        ]

    def to_database_url(self, database):
        return 'duckdb:///:memory:'

    def create_engine(self, database=None):
        credential = self.credential
        csv_path = os.path.abspath(credential.get('path'))
        if not os.path.exists(csv_path):
            raise PipeRiderDataBaseConnectionError(self.name, self.type_name, db_path=csv_path)
        table_name = self._formalize_table_name(splitext(basename(csv_path))[0])
        engine = super().create_engine(database)
        # Load csv file as table
        sql_query = f"CREATE TABLE '{table_name}' AS SELECT * FROM read_csv_auto('{csv_path}')"
        try:
            with engine.connect() as conn:
                trans = conn.begin()
                conn.execute(text('SET enable_progress_bar=true;'))
                conn.execute(text(sql_query))
                conn.execute(text('SET enable_progress_bar=false;'))
                trans.commit()
        except Exception as e:
            if isinstance(e, DBAPIError) and e.args[0].endswith('Invalid Error: String value is not valid UTF8'):
                encoding_detection = self.detect_file_encoding(csv_path)
                raise PipeRiderDataBaseEncodingError(csv_path, 'csv', encoding_detection['encoding'], 'UTF-8')
            raise e
        return engine

    def detect_file_encoding(self, file):
        import chardet
        with open(file, "rb") as fd:
            rawdata = fd.read()
            return chardet.detect(rawdata)


class ParquetDataSource(DuckDBDataSource):

    def __init__(self, name, **kwargs):
        super(DuckDBDataSource, self).__init__(name, 'parquet', **kwargs)

        self.credential_source = 'config'
        self.fields = [
            TextField('path',
                      description='Path of the Parquet file. (Support: file path, http and s3)',
                      validate=_parquet_path_validate_func),
        ]

    def to_database_url(self, database):
        return 'duckdb:///:memory:'

    def _extract_parquet_path(self):
        credential = self.credential
        parquet_path = credential.get('path')

        if _parquet_path_validate_func is False:
            raise PipeRiderDataBaseConnectionError(self.name, self.type_name, db_path=parquet_path)

        if parquet_path.startswith('https://') \
            or parquet_path.startswith('http://'):
            return 'http', parquet_path

        if parquet_path.startswith('s3://'):
            aws_access_key_id, aws_secret_access_key = get_aws_credentials()
            if aws_access_key_id is None or aws_secret_access_key is None:
                raise AwsCredentialsError('No AWS credentials found')
            s3_uri = urlparse(parquet_path)
            s3_bucket = s3_uri.netloc
            s3_region = get_s3_bucket_region(s3_bucket)
            if not s3_region:
                raise AwsUnExistedS3Bucket(s3_bucket)
            return 's3', parquet_path

        return 'file', os.path.abspath(parquet_path)

    def create_engine(self, database=None):
        type, parquet_path = self._extract_parquet_path()
        engine = super().create_engine(database)

        with engine.connect() as conn:
            trans = conn.begin()

            if type == 'http':
                # HTTP
                conn.execute(text('INSTALL httpfs'))
                conn.execute(text('LOAD httpfs'))
                url = urlparse(parquet_path)
                table_name = self._formalize_table_name(splitext(os.path.basename(url.path))[0])
                pass
            elif type == 's3':
                # S3
                s3_uri = urlparse(parquet_path)
                conn.execute(text('INSTALL httpfs'))
                conn.execute(text('LOAD httpfs'))
                aws_access_key_id, aws_secret_access_key = get_aws_credentials()
                conn.execute(text(f'set s3_access_key_id="{aws_access_key_id}"'))
                conn.execute(text(f'set s3_secret_access_key="{aws_secret_access_key}"'))
                conn.execute(text(f'set s3_region="{get_s3_bucket_region(s3_uri.netloc)}"'))
                table_name = self._formalize_table_name(splitext(os.path.basename(s3_uri.path))[0])
            else:
                # File
                table_name = self._formalize_table_name(splitext(basename(parquet_path))[0])

            sql_query = f"CREATE TABLE '{table_name}' AS SELECT * FROM read_parquet('{parquet_path}')"
            conn.execute(text('SET enable_progress_bar=true;'))
            conn.execute(text(sql_query))
            conn.execute(text('SET enable_progress_bar=false;'))
            trans.commit()
        return engine



def compare_with_recipe(**kwargs):
    """
    Generate the comparison report for your branch.
    """

    from piperider_cli.cli_utils import DbtUtil
    from piperider_cli.cli_utils.cloud import CloudConnectorHelper
    from piperider_cli.configuration import FileSystem, is_piperider_workspace_exist
    from piperider_cli.error import DbtProjectNotFoundError, RecipeConfigException
    from piperider_cli.initializer import Initializer
    from piperider_cli.recipes import RecipeConfiguration, configure_recipe_execution_flags, is_recipe_dry_run

    recipe = kwargs.get('recipe')
    summary_file = kwargs.get('summary_file')
    force_upload = kwargs.get('upload')
    enable_share = kwargs.get('share')
    open_report = kwargs.get('open')
    project_name = kwargs.get('project')
    debug = kwargs.get('debug', False)
    select = kwargs.get('select')
    modified = kwargs.get('modified')

    base_branch = kwargs.get('base_branch')

    # reconfigure recipe global flags
    configure_recipe_execution_flags(dry_run=kwargs.get('dry_run'), interactive=kwargs.get('interactive'))

    if enable_share:
        force_upload = True

    if force_upload is True and CloudConnectorHelper.is_login() is False:
        raise RecipeConfigException(
            message='Please login to PipeRider Cloud first.',
            hint='Run "piperider cloud login" to login to PipeRider Cloud.'
        )

    # Search dbt project config files
    dbt_project_dir = kwargs.get('dbt_project_dir')
    no_auto_search = kwargs.get('no_auto_search')
    dbt_project_path = DbtUtil.get_dbt_project_path(dbt_project_dir, no_auto_search, recursive=False)
    dbt_profiles_dir = kwargs.get('dbt_profiles_dir')
    if dbt_project_path:
        FileSystem.set_working_directory(dbt_project_path)
        # Only run initializer when dbt project path is provided
        Initializer.exec(dbt_project_path=dbt_project_path, dbt_profiles_dir=dbt_profiles_dir, interactive=False)
    elif is_piperider_workspace_exist() is False:
        raise DbtProjectNotFoundError()

    ret = 0
    try:
        # note: dry-run and interactive are set by configure_recipe_execution_flags
        from piperider_cli.recipe_executor import RecipeExecutor
        recipe_config: RecipeConfiguration = RecipeExecutor.exec(
            recipe_name=recipe,
            select=select,
            modified=modified,
            base_branch=base_branch,
            debug=debug)
        last = False
        base = target = None
        if not recipe_config.base.is_file_specified() and not recipe_config.target.is_file_specified():
            last = True
        else:
            base = recipe_config.base.get_run_report()
            target = recipe_config.target.get_run_report()

        if not is_recipe_dry_run():
            from piperider_cli.compare_report import CompareReport
            CompareReport.exec(a=base, b=target, last=last, datasource=None,
                               output=kwargs.get('output'), tables_from="all",
                               summary_file=summary_file,
                               force_upload=force_upload,
                               enable_share=enable_share,
                               open_report=open_report,
                               project_name=project_name,
                               show_progress=True,
                               debug=debug)
    except Exception as e:
        raise e

    return ret



import json
import os

import pytest
from piperider_cli.dbt import dbt_version, disable_dbt_compile_stats
from piperider_cli.dbt.list_task import load_manifest
from packaging import version


def _load_manifest(file_name):
    file_path = os.path.join(os.path.dirname(__file__), 'mock_dbt_data', 'compatible', file_name)
    with open(file_path, 'r') as file:
        content = file.read()
    return json.loads(content)


@pytest.fixture()
def manifest_from_1_3():
    return _load_manifest('dbt-postgres-1.3.4-manifest.json')


@pytest.fixture()
def manifest_from_1_4():
    return _load_manifest('dbt-duckdb-1.4.2-manifest.json')


@pytest.fixture()
def manifest_from_1_5():
    return _load_manifest('dbt-duckdb-1.5.1-manifest.json')


@pytest.fixture()
def manifest_from_1_6():
    return _load_manifest('dbt-duckdb-1.6.0-manifest.json')


def check_manifest_type(m):
    from dbt.contracts.graph.manifest import Manifest, WritableManifest
    assert isinstance(m, Manifest) or isinstance(m, WritableManifest)


def test_load_manifest_v7(manifest_from_1_3):
    m = load_manifest(manifest_from_1_3)
    check_manifest_type(m)


@pytest.mark.skipif(dbt_version < version.parse('v1.4'), reason='skip manifest test before dbt-core 1.4')
def test_load_manifest_v8(manifest_from_1_4):
    m = load_manifest(manifest_from_1_4)
    check_manifest_type(m)


@pytest.mark.skipif(dbt_version < version.parse('v1.5'), reason='skip manifest test before dbt-core 1.5')
def test_load_manifest_v9(manifest_from_1_5):
    m = load_manifest(manifest_from_1_5)
    check_manifest_type(m)


@pytest.mark.skipif(dbt_version < version.parse('v1.6'), reason='skip manifest test before dbt-core 1.6')
def test_load_manifest_10(manifest_from_1_6):
    m = load_manifest(manifest_from_1_6)
    check_manifest_type(m)


def test_log_functions():
    with disable_dbt_compile_stats():
        # make sure this not breaking in all dbt versions
        pass



import os
import sys
import traceback
import typing as t

import sentry_sdk
from click.core import Command, Context, Group
from rich.console import Console
from rich.markup import escape

from piperider_cli import event
from piperider_cli.error import PipeRiderError
from piperider_cli.guide import Guide

_enable_trackback: bool = os.environ.get('PIPERIDER_PRINT_TRACKBACK') == '1'

guide = Guide()

beta_message = '(Experimental)'


class BetaGroup(Group):
    def __init__(self, *args, **kwargs):
        kwargs['hidden'] = not os.environ.get('PIPERIDER_BETA', 'false').lower() == 'true'
        if kwargs.get('short_help'):
            kwargs['short_help'] = f'{beta_message} {kwargs["short_help"]}'
        super().__init__(*args, **kwargs)


class TrackCommand(Command):

    def __init__(
        self,
        name: t.Optional[str],
        context_settings: t.Optional[t.Dict[str, t.Any]] = None,
        callback: t.Optional[t.Callable[..., t.Any]] = None,
        params: t.Any = None,
        help: t.Optional[str] = None,
        epilog: t.Optional[str] = None,
        short_help: t.Optional[str] = None,
        options_metavar: t.Optional[str] = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool = False,
        beta: bool = False,
    ) -> None:
        if beta:
            short_help = f'{beta_message} {short_help}'
        super(TrackCommand, self).__init__(name, context_settings, callback, params, help, epilog, short_help,
                                           options_metavar, add_help_option, no_args_is_help, hidden, deprecated)

    def _show_error_message(self, msg, params):
        console = Console()
        if params.get('debug'):
            console.print_exception(show_locals=True)
        else:
            console.print('[bold red]Error:[/bold red] ', end='')
            console.out(msg, highlight=False)

    def _show_hint_message(self, hint):
        console = Console()
        console.print(f'[bold yellow]Hint[/bold yellow]:\n  {escape(hint)}')

    def invoke(self, ctx: Context) -> t.Any:
        status = False
        try:
            self._apply_project_parameters(ctx)

            ret = super(TrackCommand, self).invoke(ctx)
            if ret is None or ret == 0:
                guide.show_tips(ctx.command.name)
                status = True
            return ret
        except Exception as e:
            if _enable_trackback:
                print(traceback.format_exc())

            if isinstance(e, PipeRiderError):
                if e.hint:
                    self._show_hint_message(e.hint)

            ignored = False
            if isinstance(e, EOFError):
                ignored = True
                self._show_error_message(
                    f"The '{ctx.command.name}' command could not be executed "
                    f"due to the unavailability of STDIN, which is required to receive user input.", ctx.params)

                if ctx.command.name in ['login', 'select_project']:
                    self._show_hint_message("""
Consider using the following options to select a project:

    - To automatically select the first project, use `--no-interaction` option.
    - To select a specific project, use the `--project` option by the project name.
                    """.strip())
            else:
                ignored = False
                self._show_error_message(e, ctx.params)

            if not ignored:
                sentry_sdk.capture_exception(e)
            sentry_sdk.flush()
            sys.exit(1)
        finally:
            event.log_usage_event(ctx.command.name, ctx.params, status)
            event.flush_events(ctx.command.name)

    def _apply_project_parameters(self, ctx):
        # sc-30888 use PIPERIDER_API_PROJECT if `--project` did not set
        if 'project' in ctx.params and ctx.params.get('project') is None:
            ctx.params['project'] = os.environ.get('PIPERIDER_API_PROJECT')



import json
import os

from rich.console import Console

from piperider_cli import raise_exception_when_directory_not_writable, get_run_json_path
from piperider_cli.assertion_engine import AssertionEngine
from piperider_cli.configuration import Configuration
from piperider_cli.error import PipeRiderNoProfilingResultError

console = Console()


def _validate_input_result(result):
    for f in ['tables', 'id', 'created_at', 'datasource']:
        if f not in result:
            return False
    return True


class AssertionGenerator():
    @staticmethod
    def exec(input_path=None, report_dir: str = None, no_recommend: bool = False, table: str = None):
        console.rule('Deprecated', style='bold red')
        console.print(
            'Assertions Generator is deprecated and will be removed in the future. If you have a strong need for assertions, please contact us by "piperider feedback".\n')
        filesystem = Configuration.instance().activate_report_directory(report_dir=report_dir)
        raise_exception_when_directory_not_writable(report_dir)

        run_json_path = get_run_json_path(filesystem.get_output_dir(), input_path)
        if not os.path.isfile(run_json_path):
            raise PipeRiderNoProfilingResultError(run_json_path)

        with open(run_json_path) as f:
            profiling_result = json.loads(f.read())
        if not _validate_input_result(profiling_result):
            console.print(f'[bold red]Error: {run_json_path} is invalid[/bold red]')
            return 1
        console.print(f'[bold dark_orange]Generating recommended assertions from:[/bold dark_orange] {run_json_path}')

        if table:
            # only keep the profiling result of the specified table
            profiling_result['tables'] = {k: v for k, v in profiling_result['tables'].items() if k == table}
            if not profiling_result['tables']:
                console.print(f'[bold red]Error: {table} is not found from {run_json_path}[/bold red]')
                return 1

        assertion_engine = AssertionEngine(None)
        if no_recommend:
            template_assertions = assertion_engine.generate_template_assertions(profiling_result)

            # Show the assertion template files
            console.rule('Generated Assertions Templates')
            for f in template_assertions:
                console.print(f'[bold green]Assertion Templates[/bold green]: {f}')
        else:
            # Generate recommended assertions
            assertion_engine.load_assertions(profiler_result=profiling_result)
            recommended_assertions = assertion_engine.generate_recommended_assertions(profiling_result)

            # Show the recommended assertions files
            console.rule('Generated Recommended Assertions')
            for f in recommended_assertions:
                console.print(f'[bold green]Recommended Assertion[/bold green]: {f}')
